/*
 * This code is copyrighted 2002 by Reimer Mellin. You can use and modify it
 * freely, as long as derivative work credits the original author.
 * I am interested to hear about applications of this work. Please drop me a line
 * at mellinr@yahoo.com
 * $Id: WOStart.java,v 1.2 2002/08/09 20:14:22 reimer Exp $
 */

/*
 * $Log: WOStart.java,v $
 * Revision 1.2  2002/08/09 20:14:22  reimer
 * beta version
 *
 */
 
import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;


/**
 * This class is a replacement for the startup scripts WOF normally requires.<br> It offers the following functionality:<ul>
 * <li>determination of Application dir (*.woa): <ul> <li>find dir in current directory path <li>
 * find dir in the path of the WOStart.class </ul> <li>determination of $NEXT_ROOT: <ul> <li> WORootDirectory java property
 * <li>Existence of one the following directories: in one of the file system Roots determined by <code>File.listRoots()</code>:
 * <ul> <li> ROOT/Apple <li>ROOT/opt/Apple <li>ROOT/var/Apple <li>ROOT/usr/Apple <li>ROOT/System </ul> </ul>
 * <li>determination of home directory of user: <ul> <li>java user.dir property <li>java.io.tmpdir property
 * <li>user.dir property </ul> <li>scanning of the *.woa directory tree for any contained *.jar and *.zip files, adding them
 * to the CLASSPATH <li> reading and resolution of any explicitly referenced framework in <code>Contents/Config/UNIX/
 * UNIXClassPath.txt</code> <li>reading of (optional) Application class name from config file
 * <li>explicit load of Application main Class and invocation of <code>static void main(String[])</code> function </ul> TODO:
 * <ul> <li>adding the APPROOT/Contents/Java directory to CLASSPATH </ul>
 * @author Reimer Mellin
 * @version 1.0
 */
public class WOStart extends URLClassLoader {
    /*
     * WOF specific constants
     */

    private static final String[] WOHOMENAME_ARRAY = {"Apple",
        											"/opt/Apple",
                                                    "/var/Apple",
                                                    "/usr/Apple",
                                                    "/System"};
    private static final String WOLOCALDIR_KEY = "Local";
    private static final String WOA_KEY = ".woa";
    private static final String WOHOME_KEY = "WORootDirectory"; // set to define NEXT_ROOT
    private static final String WOUSER_KEY = "WOUserDirectory"; // set to app root
    private static final String WOCURRENTDIR_KEY = "webobjects.user.dir"; // nsbundle prefers that


    /*
     * Our Conventions
     */

    private static final String[] JAVA_ARCHIVE_EXTENSIONS = { ".jar", ".zip" };

    /*
     * Token for config files
     */

    private static final String APPROOT_TOKEN = "APPROOT";
    private static final String WOROOT_TOKEN = "WOROOT";
    private static final String HOMEROOT_TOKEN = "HOMEROOT";
    private static final String LOCALROOT_TOKEN = "LOCALROOT";

    /*
     * Keys for java System properties
     */

    private static final String CURRENTDIR_KEY = "user.dir";
    private static final String TMPDIR_KEY = "java.io.tmpdir";
    private static final String USERHOME_KEY = "user.home";
    private static final String CLASSPATH_KEY = "java.class.path";

    /*
     * Starting of WOFApp related variables
     */

    private File 	appWoa;
    private File 	appRoot;
    private File 	woRoot;
    private File 	homeRoot;
    private File 	localRoot;
    private String 	applicationClassName = "Application";
    private Set 	fileSet;

    /*
     * ClassLoader specific variables
     */
    private Hashtable	patchTable = null;

    /*
     * Constructors
     */
	public WOStart(URL[] urls, ClassLoader parent) {
        super( urls, parent);
    }

	public WOStart(URL[] urls) {
        super(urls);
    }

    public WOStart(URL[] urls,
                      ClassLoader parent,
                      URLStreamHandlerFactory factory) {
		super( urls, parent, factory);
    }

	/*
     * ClassLoader specific functions
     */

    private void addPatchDirectory(File patchDirectory) {
        HashSet 	aSet = new HashSet();
		Iterator	anIterator;
        String		prefixPath = patchDirectory.getPath();

        if( this.patchTable == null) {
            this.patchTable = new Hashtable();
        }

        this.addFilesToSet( patchDirectory, aSet, new String[] {".class"});
		anIterator = aSet.iterator();

		while( anIterator.hasNext()) {
			File 	aPatchFile = (File)anIterator.next();
            String	patchPath = aPatchFile.getPath();
			int  	prefixOffset = prefixPath.length()+1; // jump '/'
            int  	postfixOffset;
            String 	aKey;

            postfixOffset = patchPath.lastIndexOf( ".class");
			aKey = patchPath.substring( prefixOffset, postfixOffset);
            aKey = aKey.replace( File.separatorChar, '.');

            if(this.patchTable.containsKey(aKey)) {
				System.err.println("Already contains patchs -- ignoring " + aPatchFile);
            } else {
                this.patchTable.put(aKey, aPatchFile);
            }
        }
    }

    /*
     * Patch Facility
     */
	public Class loadClass(String name)
                throws ClassNotFoundException {
        if( (this.patchTable != null) && (this.patchTable.containsKey(name))) {
			/*
             * ok, load the patch class instead
             */
            byte[]	someBytes = null;
			File	aFile = (File)this.patchTable.get(name);
			try
			{
				FileInputStream		aFileInputStream = new FileInputStream( aFile );
				BufferedInputStream	aBufferedStream = new BufferedInputStream( aFileInputStream);
                int					length = (int) aFile.length();

				someBytes = new byte[ length ];

				aBufferedStream.read(someBytes, 0,  length );
				aBufferedStream.close();

                return this.defineClass( name, someBytes, 0, length);
			}
			catch(IOException anException)
			{
                throw new ClassNotFoundException("Unable to load patch for " + name + " " + anException);
			}
		}
        return super.loadClass(name);
    }

    /*
     * Static Methods
     */

    public static void main(String[] argv) {
        URL[]	urls = new URL[] {};
        WOStart mainObject = new WOStart(urls);
        if (mainObject.getAppRoot() == null) {
            System.err.println("Sorry can not find Application directory. Unable to proceed.");
            System.err.println("current directory or WOStart.class file should be in or below *.woa directory.");
            System.exit(1);
        }
        mainObject.addPatchDirectory(new File(mainObject.getAppRoot(), "Patches"));
        mainObject.processWoa();
        mainObject.processConfigFile();
        mainObject.invokeApplication(argv);
    }

    /*
     * walks up the path and looks for directory ending in *.woa
     */

    private File findWoa(File aFile) {
        String aFileName;
        if (aFile == null) {
            return aFile;
        }
        aFileName = aFile.getName();
        if (aFileName.endsWith(WOStart.WOA_KEY)) {
            return aFile;
        }
        return this.findWoa(aFile.getParentFile());
    }

    /*
     * Accessors for machine/runtime dependent file-dirs
     */

    public File getAppWoa() {
        if (this.appWoa == null) {
            // start with user.dir ... find *.woa in path
            // if that fails ->path to own class
            File aFile = null;


            aFile = new File(System.getProperty(WOStart.WOUSER_KEY));
            aFile = this.findWoa(aFile);

            if(aFile == null) {
            	aFile = new File(System.getProperty(WOStart.CURRENTDIR_KEY));
            	aFile = this.findWoa(aFile);
            }

            if (aFile == null) {
                Class aClass = this.getClass();
                String aClassName = aClass.getName();
                String aResourceName = "/" + aClassName.replace('.', '/') + ".class";
                java.net.URL anURL = aClass.getResource(aResourceName);
                aFile = new File(anURL.getPath());
            	aFile = this.findWoa(aFile);
            }
            if (aFile == null) {
                return null;
            }
            this.appWoa = aFile;
        }
        return this.appWoa;
    }

	public File getAppRoot() {
        if (this.appRoot == null) {
            File aFile = this.getAppWoa();

        	if(aFile == null) {
				return null;
        	}
        	this.appRoot = new File(aFile, "Contents");
        }
        return this.appRoot;
    }


    /*
     * if The NEXT_ROOT is not specified as a java property, then
     * apply a heuristic to find the directory:
     *   - traverse all root file systems
     *   - check for existience of /Apple,/opt/Apple,/System etc..
     * if found, declare victory and move on
     * if not...oh well
     */
    public File getWoRoot() {
        if (this.woRoot == null) {
            // start with wo.root
            File aFile = null;
            if (System.getProperty(WOStart.WOHOME_KEY) != null) {
                aFile = new File(System.getProperty(WOStart.WOHOME_KEY));
            }
            if (aFile == null) {
                File[] rootList = File.listRoots();
                for (int i = 0; i < rootList.length; i++) {
                    for (int j = 0; j < WOStart.WOHOMENAME_ARRAY.length; j++) {
                        aFile = new File(rootList[i], WOStart.WOHOMENAME_ARRAY[j]);
                        if (aFile.isDirectory()) {
                            break;
                        } else {
                            aFile = null;
                        }
                    }
                    if (aFile != null) {
                        break;
                    }
                    aFile = null;
                }
            }
            if (aFile != null) {
                this.woRoot = aFile;
            } else {
                return null;
            }
        }
        return this.woRoot;
    }

    public File getLocalRoot() {
        if (this.localRoot == null) {
            // woroot/Local
            this.localRoot = new File(this.getWoRoot(), WOStart.WOLOCALDIR_KEY);
        }
        return this.localRoot;
    }

    public File getHomeRoot() {
        if (this.homeRoot == null) {
            // user.home
            // if not found java.io.tmpdir
            // if not found top of user.dir
            File aFile = null;
            if (System.getProperty(WOStart.USERHOME_KEY) != null) {
                aFile = new File(System.getProperty(WOStart.USERHOME_KEY));
            } else if (System.getProperty(WOStart.TMPDIR_KEY) != null) {
                aFile = new File(System.getProperty(WOStart.TMPDIR_KEY));
            } else if (System.getProperty(WOStart.CURRENTDIR_KEY) != null) {
                aFile = new File(System.getProperty(WOStart.CURRENTDIR_KEY));
            }
            if (aFile != null) {
                this.homeRoot = aFile;
            } else {
                return null;
            }
        }
        return this.homeRoot;
    }

    /*
     * private helper function to find all *.x files and add them to a set
     */

    private void addFilesToSet(File aFile, Set aSet, String[] someExtensions) {
        if (aFile.isDirectory()) {
            File[] 	allFiles = aFile.listFiles();

            for (int i = 0; i < allFiles.length; i++) {
                this.addFilesToSet(allFiles[i], aSet, someExtensions);
            }
        } else {
            for (int i = 0; i < someExtensions.length; i++) {
                if (aFile.getName().endsWith(someExtensions[i])) {
                    try {
                        aSet.add(aFile);
                    } catch (Exception anException) {
                        System.err.println(anException);
                    }
                    return;
                }
            }
        }
    }

    /*
     * processing the woa application wrapper and assemble all fileURLs relevant for
     * the classpath (jar, zip) into the fileSet
     */

    public void processWoa() {
        File 	aFile = this.getAppWoa();

        this.fileSet = new HashSet();
        this.addFilesToSet(aFile, this.fileSet, WOStart.JAVA_ARCHIVE_EXTENSIONS);
    }

	/*
     * scan the config file and extract framework paths and
     * main class name
     */

    private void processConfigFileLine(String aLine) {
        File aFile = null;

		/*
         * ugly ... so shoot me
         */

        if (aLine.startsWith(WOStart.APPROOT_TOKEN)) {
            aFile = new File(this.getAppRoot(), aLine.substring(WOStart.APPROOT_TOKEN.length()));
        } else if (aLine.startsWith(WOStart.WOROOT_TOKEN)) {
            aFile = new File(this.getWoRoot(), aLine.substring(WOStart.WOROOT_TOKEN.length()));
        } else if (aLine.startsWith(WOStart.HOMEROOT_TOKEN)) {
            aFile = new File(this.getHomeRoot(), aLine.substring(WOStart.HOMEROOT_TOKEN.length()));
        } else if (aLine.startsWith(WOStart.LOCALROOT_TOKEN)) {
            aFile = new File(this.getLocalRoot(), aLine.substring(WOStart.LOCALROOT_TOKEN.length()));
        } else if (aLine.indexOf("ApplicationClass") > 0) {
            int i = aLine.indexOf("==");
            if (i > 0) {
                String aClassName = aLine.substring(i + 2); // size of '=='
                this.applicationClassName = aClassName.trim();
            }
        }
        if (aFile != null) {
            try {
                this.fileSet.add(aFile);
            } catch (Exception anException) {
                System.err.println(anException);
            }
        }
    }

    /*
     * trying to find the generated config file, replacing the
     * TOKENS with actual prefices and adding the resulting File's into a set
     */

    public void processConfigFile() {
        String configFilePath = this.getAppRoot() + File.separator + "UNIX" + File.separator + "UNIXClassPath.txt";
        File configFile = new File(configFilePath);
        if (configFile.exists()) {
            try {
                BufferedReader aBuffer = new BufferedReader(new InputStreamReader(new FileInputStream(configFile)));
                while (true) {
                    try {
                        this.processConfigFileLine(aBuffer.readLine());
                    } catch (Exception anException) {
                        break; // EOF
                    }
                }
                aBuffer.close();
            } catch (Exception anException) {
                System.err.println("Can not open config File -- ignoring");
            }
        } else {
            System.err.println("Can not find configuration file -- ignoring " + configFile);
        }
    }

    /*
     * main function, builds a URL array for ClassLoader and also build a
     * new class path property. With the property NSBundle will fail miserably.
     * load the application main class and invoke main().
     */

    public void invokeApplication(String[] argv) {
        Class 	mainClass = null;
        Method 	aMethod = null;

        try {
            Object[] 		fileArray = fileSet.toArray();
            StringBuffer	aBuffer = new StringBuffer();

            aBuffer.append(System.getProperty(WOStart.CLASSPATH_KEY));

            for (int i= fileSet.toArray().length ;i > 0; i--) {
                File	aFile = (File)fileArray[i-1];

                aBuffer.append(File.pathSeparator);
                aBuffer.append(aFile.getPath());

                this.addURL(aFile.toURL());
            }

            /*
             * NSBundle initialization does not work without a CLASSPATH property.
             * Thats due to some JDK1.1'isms in the code.
             * They could ask the Classes for their origin and then derive
             * all the necessary info from that, but alas ....
             * WOCURRENTDIR is also needed due to some brain damage from NSBundle:
             * if the CLASSPATH contains '.' and user.dir is the *.woa then NSBundle
             * throws an exception....
             * The WOOpenURL functionality requires the NEXT_ROOT to be defined in WOHOME
             */

            System.setProperty(WOStart.CLASSPATH_KEY, aBuffer.toString());
			System.setProperty(WOStart.WOCURRENTDIR_KEY, this.getAppRoot().getPath());
			System.setProperty(WOStart.WOUSER_KEY, this.getAppRoot().getPath());
            System.setProperty(WOStart.WOHOME_KEY, this.getWoRoot().getPath()); // needed for WOOpenURL

			/*
             * load the Application class
             * any dependencies on WOF are being resolved automagically
			 * find any static main method and invoke with arguments
             */
            mainClass = this.loadClass(this.applicationClassName);
            aMethod = mainClass.getDeclaredMethod("main",
                new Class[] { String[].class });
            aMethod.invoke(mainClass,
                new Object[] { argv });
        } catch (Exception anException) {
            System.err.println("Can not find or invoke WOF application class");
            System.err.println(anException);
            System.exit(1);
        }
    }
}